<?php
/*******************************************************************************
 *	BGpx.php		Simple GPX file parser
 *	Copyright		Terry Barnaby,	BeamWeb,	2020-03-29
 *******************************************************************************
 */
require_once('simple_html_dom.php');

class BGpxPoint {
	function __construct($long, $lat, $h = 0){
		$this->long = $long;
		$this->lat = $lat;
		$this->h = $h;
	}
	function __toString(){
		return "{$this->long}, {$this->lat}, {$this->h}";
	}
};

class BGpx {
	function __construct(){
		$this->first = 1;
		$this->longMax = 0;
		$this->longMin = 0;
		$this->latMax = 0;
		$this->latMin = 0;
		$this->start = new BGpxPoint(0, 0, 0);
		$this->lastPos = Null;
		$this->points = array();
		$this->lastRoughPos = Null;
		$this->pointsRough = array();
		$this->title = "";
		$this->distance = 0.0;
		$this->climb = 0.0;
		$this->effort = 0.0;
		$this->maxGradient = 0.0;
		$this->rad = 6378137.0; // Radius of earth in metres
		$this->changeHeights = 0;
		$this->minDistance = 10.0;
		$this->roughDistance = 1000.0;
		$this->heightFilter = 0.0;
		$this->heightFilterFactor = 0.95;
		$this->ret = "";
		$this->url = "";
		$this->copyright = "";
	}
	
	function setConfig($changeHeights){
		$this->changeHeights = $changeHeights;
	}
	
	function setInfo($url, $copyright){
		$this->url = $url;
		$this->copyright = $copyright;
	}
	
	function parse($filename){
		$this->title = "";
		$this->distance = 0.0;

		if(!file_exists($filename))
			return 1;
		
		$this->first = 1;

		$xml = simplexml_load_file($filename);

		foreach($xml->children() as $child){
			$tag = $child->getName();
			//print("Do: {$tag}\n");

			if($tag == "rte"){
				foreach($child->children() as $grandchild){
					$tag = $grandchild->getName();
					if($tag == "name"){
						//print("Name: {$grandchild}<br/>\n");
						$this->title = $grandchild;
					}
					else if($tag == 'rtept'){
						//print($grandchild["lat"] . " " . $grandchild["lon"] . "\n");
						if(isset($grandchild->ele))
							$ele = floatval($grandchild->ele);
						else
							$ele = 0.0;

						$p = new BGpxPoint(floatval($grandchild["lon"]), floatval($grandchild["lat"]), $ele);
						$this->addPoint($p);
					}
				}
			}
			
			// Tracks
			else if($tag == 'trk'){
				foreach($child->children() as $grandchild){
					$grandname = $grandchild->getName();
					if($grandname == "name"){
						//print("Name: {$grandchild}<br/>\n");
						$this->title = $grandchild;
					}
					else if($grandname == "trkseg"){
						foreach($grandchild->children() AS $greatgrandchild){
							$greatgrandname = $greatgrandchild->getName();
							//print_r($greatgrandname . "  ");
							if($greatgrandname == "trkpt"){
								//print($greatgrandchild["lat"] . " " . $greatgrandchild["lon"] . "<br/>\n");
								//print("Element:"); print_r($greatgrandchild);
								if(isset($greatgrandchild->ele))
									$ele = floatval($greatgrandchild->ele);
								else
									$ele = 0.0;

								$p = new BGpxPoint(floatval($greatgrandchild["lon"]), floatval($greatgrandchild["lat"]), $ele);
								$this->addPoint($p);

								foreach( $greatgrandchild->children() as $c){
									//print($c . "<br/>\n");
									if($c->getname() == "ele"){
										//print_r($c);
									}
								}
							}
						}
					}
				}
			} 
			
			// This wpt code might not work !!!
			else if(0 and ($tag == "wpt")){
				//print_r($tag . "    ");
				//print($child["lat"] . " " . $child["lon"] . "\n");
				if(isset($child->ele))
					$ele = floatval($child->ele);
				else
					$ele = 0.0;

				$p = new BGpxPoint(floatval($child["lon"]), floatval($child["lat"]), $ele);
				$this->addPoint($p);
				
				$tag = $child->children()->getName();
				if($tag == "ele"){
					//print($child->children() . "<br/>\n");
				}
			}

		}
		return 0;
	}
	
	function save($filename, $title){
		global $sys;
		
		$file = fopen($filename, "w");
		fwrite($file, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		fwrite($file, "<gpx xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.topografix.com/GPX/1/1\" xmlns:gpxdata=\"http://www.cluetrust.com/XML/GPXDATA/1/0\" xsi:schemaLocation=\"http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd http://www.cluetrust.com/XML/GPXDATA/1/0 http://www.cluetrust.com/Schemas/gpxdata10.xsd\" version=\"1.1\" creator=\"http://{$sys->baseUrl}\">\n");

		fwrite($file, "<metadata>\n");
		fwrite($file, "<name>{$title}</name>\n");

		if($this->copyright != "")
			fwrite($file, "<copyright author=\"{$this->copyright}\"></copyright>\n");

		fwrite($file, "<link href=\"{$this->url}\"><text>{$title}</text></link>\n");
		
		fwrite($file, "</metadata>\n");
		fwrite($file, "<trk>\n");
		fwrite($file, "<name>{$title}</name>\n");
		fwrite($file, "<trkseg>\n");
		foreach($this->points as $p){
			fwrite($file, "<trkpt lat=\"{$p->lat}\" lon=\"{$p->long}\"><ele>{$p->h}</ele></trkpt>\n");
		}
		
		fwrite($file, "</trkseg>\n");
		fwrite($file, "</trk>\n");
		fwrite($file, "</gpx>\n");
		fclose($file);
	}
	
	function addPoint($p){
		if($this->first){
			if($this->changeHeights){
				try {
					$p->h = mapGetHeight($p->lat, $p->long);
				}
				catch(Exception $e){
					$this->ret = "Error: No height map";
					$this->changeHeights = 0;
				}
			}

			$p->hf = $p->h;
			$this->longMax = $this->longMax = $p->long;
			$this->latMax = $this->latMin = $p->lat;
			$this->start = $p;
			$this->lastPos = $p;
			$this->points[] = $p;
			$this->lastRoughPos = p;
			$this->pointsRough[] = $p;
			$this->first = 0;
		}
		else {
			// Only consider points a reasonable distance apart
			if($this->calcDistance($this->lastPos, $p) > $this->minDistance){
				if($this->changeHeights)
					$p->h = mapGetHeight($p->lat, $p->long);

				$this->heightFilter = ($this->heightFilterFactor * $this->heightFilter) + ((1.0 - $this->heightFilterFactor) * $p->h);
				$p->hf = $this->heightFilter;

				$this->distance += $this->calcDistance($this->lastPos, $p);
				$this->climb += $this->calcClimb($this->lastPos, $p);
				$this->effort += $this->calcEffort($this->lastPos, $p);
				$g = $this->calcGradient($this->lastPos, $p);
				if($g > $this->maxGradient)
					$this->maxGradient = $g;

				if($p->long > $this->longMax)
					$this->longMax = $p->long;
				if($p->long < $this->longMin)
					$this->longMin = $p->long;
				if($p->lat > $this->latMax)
					$this->latMax = $p->lat;
				if($p->lat < $this->latMin)
					$this->latMin = $p->lat;

				if(0){
					$d = $this->calcDistance($this->lastPos, $p);
					$c = $this->calcClimb($this->lastPos, $p);
					print("Point: {$this->distance} {$p->long} {$p->lat} {$p->h}: {$d}, {$c}, {$g}: {$this->maxGradient}\n"); 
				}
				
				$this->lastPos = $p;
				$this->points[] = $p;
				if($this->calcDistance($this->lastRoughPos, $p) > $this->roughDistance){
					$this->lastRoughPos = $p;
					$this->pointsRough[] = $p;
				}

			}
		}

	}

	function calcDistance($p1, $p2){
		$d2r = M_PI / 180.0;
		$dlong = abs($p2->long - $p1->long) * $d2r;
		$dlat = abs($p2->lat - $p1->lat) * $d2r;
		$a = pow(sin($dlat / 2.0), 2) + cos($p1->lat * $d2r) * cos($p2->lat * $d2r) * pow(sin($dlong / 2.0), 2);

		$c = 2.0 * atan2(sqrt($a), sqrt(1.0 - $a));
		return $this->rad * $c;
	}
	
	function calcClimb($p1, $p2){
		if($p2->h > $p1->h)
			return $p2->h - $p1->h;
		else
			return 0;
	}
	
	function calcGradient($p1, $p2){
		$d = $this->calcDistance($p1, $p2);
		$z = $p2->hf - $p1->hf;
		
		if($d > 0){
			return $z / $d;
		}
		else {
			return 0;
		}
	}
	
	function calcEffort($p1, $p2){
		$d = $this->calcDistance($p1, $p2);
		$g = $this->calcGradient($p1, $p2);

		if($g > -0.02){
			return ($d * ($g + 0.02));
		}
		else {
			return 0;
		}
	}
	
	function centre(){
		return new BGpxPoint($this->longMin + ($this->longMax - $this->longMin) / 2, $this->latMin + ($this->latMax - $this->latMin) / 2);
	}

	function start(){
		return $this->start;
	}

	function distance(){
		return $this->distance;
	}

	function climb(){
		return $this->climb;
	}

	function maxGradient(){
		return $this->maxGradient;
	}

	function effort(){
		return $this->effort;
	}

	function title(){
		return $this->title;
	}
	
	function getPointsRough(){
		$str = "";
		foreach($this->pointsRough as $p){
			if($str)
				$str .= ",";
			$str .= number_format($p->long, 3) ."," . number_format($p->lat, 3);
		}
		return $str;
	}
};

// Simple tests
if(0){
	require_once "../modules/maps/SRTMGeoTIFFReader.php";
	function mapGetHeight($lat, $long){
		static $dataReader = Null;

		if($dataReader == Null)
			$dataReader = new SRTMGeoTIFFReader("../modules/maps/GeoData");

		//$h = $dataReader->getElevation($lat, $long, $interpolate = false);
		//print("mapGetHeight: {$lat} {$long} {$h}\n");
		return $dataReader->getElevation($lat, $long, $interpolate = false);
	}
	
	$sys = new stdClass();
	$sys->baseUrl = "http://tandem-club.kingnet";

	$gpx = new BGpx();
	$gpx->setConfig(1);
	
	//$r = $gpx->parse("/home/terry/tandemClub/www/www/files/rides/5/5-2020-06-22_1003 Raw.gpx");
	//$r = $gpx->parse("/home/terry/tandemClub/routes//tcride6/ChippingSodbury_Bath_Loop_5.gpx");
	$r = $gpx->parse("/home/terry/Downloads/72-tc2015-north-east-39.gpx");
	$gpx->save("test.gpx", "My Ride");

	print("Result: {$r}\n");

	print("Center: {$gpx->centre()}\n");
	print("Start: {$gpx->start()}\n");
	print("longMax: {$gpx->longMax}\n");
	print("longMin: {$gpx->longMin}\n");
	print("latMax: {$gpx->latMax}\n");
	print("latMin: {$gpx->latMin}\n");
	print("Distance: {$gpx->distance()}\n");
	print("Climb: {$gpx->climb()}\n");
	print("maxGradient: {$gpx->maxGradient()}\n");
	print("Effort: {$gpx->effort()}\n");
	$numPoints = count($gpx->points);
	print("NumPoints: {$numPoints}\n");
}

//use phpGPX\phpGPX;
if(0){
	set_include_path(".". PATH_SEPARATOR . get_include_path());
	function loadClass($class){
		$class = str_replace('\\', '/', $class) . ".php";
		print("LoadClass: {$class}\n");
		require_once($class);
	}

	spl_autoload_register("loadClass");
	//print("IncludeTree\n");
	//require_once("phpGPX/Models/Summarizable.php");
	//includeTree("phpGPX");
	//print("IncludeTree: Done\n");

	$gpx = new phpGPX();
	$file = $gpx->load("/home/terry/Downloads/72-tc2015-north-east-39.gpx");

	phpGPX::$PRETTY_PRINT = true;
	$file->save("test.gpx", phpGPX::XML_FORMAT);

	foreach($file->tracks as $track) {
		$arr = $track->stats->toArray();
		foreach($arr as $k => $v){
			print("{$k}: {$v}\n");
		}
		//var_dump($track->stats->toArray());
	}
}
?>
